@@ -56,6 +56,7 @@ This module supports Python 2.6 and later.
 
 """
 
+import errno
 import io
 import os
 import sys
@@ -86,9 +87,11 @@ except ImportError:
         # Should be safe to reuse `SystemRandom` - not software state dependant
         randbits = random.SystemRandom().getrandbits
     else:
+
         def randbits(nb):
             return random.Random().getrandbits(nb)
 
+
 try:
     import gzip
 except ImportError:
@@ -104,18 +107,31 @@ if sys.version_info[0] == 2:
 
 
 class ConcurrentRotatingFileHandler(BaseRotatingHandler):
-    """
-    Handler for logging to a set of files, which switches from one file to the
+    f"""Handler for logging to a set of files, which switches from one file to the
     next when the current file reaches a certain size. Multiple processes can
     write to the log file concurrently, but this may mean that the file will
     exceed the given size.
+
+    Version {__version__} by {__author__}.
     """
 
     def __init__(
-            self, filename, mode='a', maxBytes=0, backupCount=0,
-            encoding=None, debug=False, delay=None, use_gzip=False,
-            owner=None, chmod=None, umask=None, newline=None, terminator="\n",
-            unicode_error_policy='ignore',
+        self,
+        filename,
+        mode="a",
+        maxBytes=0,
+        backupCount=0,
+        encoding=None,
+        debug=False,
+        delay=None,
+        use_gzip=False,
+        owner=None,
+        chmod=None,
+        umask=None,
+        newline=None,
+        terminator="\n",
+        unicode_error_policy="ignore",
+        lock_file_directory=None,
     ):
         """
         Open the specified file and use it as the stream for logging.
@@ -144,6 +160,11 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
         :param unicode_error_policy: should be one of 'ignore', 'replace', 'strict'
         Determines what happens when a message is written to the log that the stream encoding
         doesn't support. Default is to ignore, i.e., drop the unusable characters.
+        :param lock_file_directory: name of directory for all lock files as alternative
+        living space; this is useful for when the main log files reside in a cloud synced
+        drive like Dropbox, OneDrive, Google Docs, etc, which may prevent the lock files
+        from working correctly. The lock file must be accessible to all processes writing
+        to a shared log, including across all different hosts (machines).
 
         By default, the file grows indefinitely. You can specify particular
         values of maxBytes and backupCount to allow the file to rollover at
@@ -189,25 +210,28 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
         self.gzip_buffer = 8096
         self.maxLockAttempts = 20
 
-        if unicode_error_policy not in ('ignore', 'replace', 'strict'):
-            unicode_error_policy = 'ignore'
+        if unicode_error_policy not in ("ignore", "replace", "strict"):
+            unicode_error_policy = "ignore"
             warnings.warn(
                 "Invalid unicode_error_policy for concurrent_log_handler: "
                 "must be ignore, replace, or strict. Defaulting to ignore.",
-                UserWarning)
+                UserWarning,
+            )
         self.unicode_error_policy = unicode_error_policy
 
         if delay not in (None, True):
             warnings.warn(
-                'concurrent_log_handler parameter `delay` is now ignored and implied as True, '
-                'please remove from your config.',
-                DeprecationWarning)
+                "concurrent_log_handler parameter `delay` is now ignored and implied as True, "
+                "please remove from your config.",
+                DeprecationWarning,
+            )
 
         # Construct the handler with the given arguments in "delayed" mode
         # because we will handle opening the file as needed. File name
         # handling is done by FileHandler since Python 2.5.
         super(ConcurrentRotatingFileHandler, self).__init__(
-            filename, mode, encoding=encoding, delay=True)
+            filename, mode, encoding=encoding, delay=True
+        )
 
         self.terminator = terminator or "\n"
 
@@ -217,14 +241,15 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
             # noinspection PyUnresolvedReferences
             self._set_gid = grp.getgrnam(self.owner[1]).gr_gid
 
-        self.lockFilename = self.getLockFilename()
+        self.lockFilename = self.getLockFilename(lock_file_directory)
         self.is_locked = False
 
-    def getLockFilename(self):
+    def getLockFilename(self, lock_file_directory):
         """
         Decide the lock filename. If the logfile is file.log, then we use `.__file.lock` and
         not `file.log.lock`. This only removes the extension if it's `*.log`.
 
+        :param lock_file_directory: name of the directory for alternative living space of lock files
         :return: the path to the lock file.
         """
         if self.baseFilename.endswith(".log"):
@@ -235,7 +260,21 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
         lock_path, lock_name = os.path.split(lock_file)
         # hide the file on Unix and generally from file completion
         lock_name = ".__" + lock_name
-        return os.path.join(lock_path, lock_name)
+        if lock_file_directory:
+            self.__create_lock_directory__(lock_file_directory)
+            return os.path.join(lock_file_directory, lock_name)
+        else:
+            return os.path.join(lock_path, lock_name)
+
+    @staticmethod
+    def __create_lock_directory__(lock_file_directory):
+        if not os.path.exists(lock_file_directory):
+            try:
+                os.makedirs(lock_file_directory)
+            except OSError as err:
+                if err.errno != errno.EEXIST:
+                    # If directory already exists, then we're done. Everything else is fishy...
+                    raise
 
     def _open_lockfile(self):
         if self.stream_lock and not self.stream_lock.closed:
@@ -243,7 +282,9 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
             return
         lock_file = self.lockFilename
         self._console_log(
-            "concurrent-log-handler %s opening %s" % (hash(self), lock_file), stack=False)
+            "concurrent-log-handler %s opening %s" % (hash(self), lock_file),
+            stack=False,
+        )
 
         with self._alter_umask():
             self.stream_lock = open(lock_file, "wb", buffering=0)
@@ -268,7 +309,11 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
         with self._alter_umask():
             # noinspection PyArgumentList
             stream = io.open(
-                self.baseFilename, mode=mode, encoding=self.encoding, newline=self.newline)
+                self.baseFilename,
+                mode=mode,
+                encoding=self.encoding,
+                newline=self.newline,
+            )
 
         self._do_chown_and_chmod(self.baseFilename)
 
@@ -287,7 +332,7 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
                 os.umask(prev_umask)
 
     def _close(self):
-        """ Close file stream.  Unlike close(), we don't tear anything down, we
+        """Close file stream.  Unlike close(), we don't tear anything down, we
         expect the log to be re-opened after rotation."""
 
         if self.stream:
@@ -304,13 +349,23 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
         if not self._debug:
             return
         import threading
+
         tid = threading.current_thread().name
         pid = os.getpid()
-        stack_str = ''
+        stack_str = ""
         if stack:
             stack_str = ":\n" + "".join(traceback.format_stack())
         asctime = time.asctime()
-        print("[%s %s %s] %s%s" % (tid, pid, asctime, msg, stack_str,))
+        print(
+            "[%s %s %s] %s%s"
+            % (
+                tid,
+                pid,
+                asctime,
+                msg,
+                stack_str,
+            )
+        )
 
     def emit(self, record):
         """
@@ -362,7 +417,7 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
                 # Try to emit in a form acceptable to the output encoding
                 # The unicode_error_policy determines whether this is lossy.
                 try:
-                    encoding = getattr(stream, 'encoding', self.encoding or 'us-ascii')
+                    encoding = getattr(stream, "encoding", self.encoding or "us-ascii")
                     msg_bin = msg.encode(encoding, self.unicode_error_policy)
                     msg = msg_bin.decode(encoding, self.unicode_error_policy)
                     stream.write(msg)
@@ -380,27 +435,28 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
         term = self.terminator
         policy = self.unicode_error_policy
 
-        encoding = getattr(stream, 'encoding', None)
+        encoding = getattr(stream, "encoding", None)
 
         # as far as I can tell, this should always be set from io.open, but just in case...
         if not encoding:
             if not self.encoding:
                 self._console_log(
-                    "Warning, unable to determine encoding of logging stream; assuming utf-8")
-            encoding = self.encoding or 'utf-8'
+                    "Warning, unable to determine encoding of logging stream; assuming utf-8"
+                )
+            encoding = self.encoding or "utf-8"
 
-        if not isinstance(msg, unicode):
-            msg = unicode(msg, encoding, policy)
+        if not isinstance(msg, unicode):  # noqa: F821
+            msg = unicode(msg, encoding, policy)  # noqa: F821
 
         # Add in the terminator.
-        if not isinstance(term, unicode):
-            term = unicode(term, encoding, policy)
+        if not isinstance(term, unicode):  # noqa: F821
+            term = unicode(term, encoding, policy)  # noqa: F821
         msg = msg + term
         stream.write(msg)
 
     def _do_lock(self):
         if self.is_locked:
-            return   # already locked... recursive?
+            return  # already locked... recursive?
         self._open_lockfile()
         if self.stream_lock:
             for i in range(self.maxLockAttempts):
@@ -413,7 +469,8 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
                     continue
             else:
                 raise RuntimeError(
-                    "Cannot acquire lock after %s attempts" % (self.maxLockAttempts,))
+                    "Cannot acquire lock after %s attempts" % (self.maxLockAttempts,)
+                )
         else:
             self._console_log("No self.stream_lock to lock", stack=True)
 
@@ -431,7 +488,7 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
 
     def close(self):
         """
-        Close log stream and stream_lock. """
+        Close log stream and stream_lock."""
         self._console_log("In close()", stack=True)
         try:
             self._close()
@@ -467,12 +524,13 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
         except (IOError, OSError):
             exc_value = sys.exc_info()[1]
             self._console_log(
-                "rename failed.  File in use? exception=%s" % (exc_value,), stack=True)
+                "rename failed.  File in use? exception=%s" % (exc_value,), stack=True
+            )
             return
 
-        gzip_ext = ''
+        gzip_ext = ""
         if self.use_gzip:
-            gzip_ext = '.gz'
+            gzip_ext = ".gz"
 
         def do_rename(source_fn, dest_fn):
             self._console_log("Rename %s -> %s" % (source_fn, dest_fn + gzip_ext))
@@ -497,9 +555,9 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
         # code complexity isn't warranted.
 
         do_renames = []
-        for i in range(1, self.backupCount - 1):
-            sfn = "%s.%d" % (self.baseFilename, i)
-            dfn = "%s.%d" % (self.baseFilename, i + 1)
+        for i in range(1, self.backupCount):
+            sfn = self.rotation_filename("%s.%d" % (self.baseFilename, i))
+            dfn = self.rotation_filename("%s.%d" % (self.baseFilename, i + 1))
             if os.path.exists(sfn + gzip_ext):
                 do_renames.append((sfn, dfn))
             else:
@@ -507,10 +565,10 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
                 # at the expected name.
                 break
 
-        for (sfn, dfn) in reversed(do_renames):
+        for sfn, dfn in reversed(do_renames):
             do_rename(sfn, dfn)
 
-        dfn = self.baseFilename + ".1"
+        dfn = self.rotation_filename(self.baseFilename + ".1")
         do_rename(tmpname, dfn)
 
         if self.use_gzip:
@@ -570,6 +628,6 @@ class ConcurrentRotatingFileHandler(BaseRotatingHandler):
 
 # Publish this class to the "logging.handlers" module so that it can be use
 # from a logging config file via logging.config.fileConfig().
-import logging.handlers
+import logging.handlers  # noqa: E402
 
 logging.handlers.ConcurrentRotatingFileHandler = ConcurrentRotatingFileHandler
